# Programmer used for In System Programming
ISP_PROG = STK200
# device the ISP programmer is connected to
ISP_DEV = lpt1
# Programmer used for serial programming (using the bootloader)
SERIAL_PROG = avr109
# device the serial programmer is connected to
SERIAL_DEV = COM1

# programs
CC = avr-gcc
OBJCOPY = avr-objcopy
OBJDUMP = avr-objdump
AS = avr-as
CP = cp
RM = rm -f
AVRDUDE = avrdude
AVRDUDE_BAUDRATE = 115200
SIZE = avr-size

all:

$(PWD)/config.mk:
	@echo "# Put your own config here!" > $@
	@echo "#SERIAL_DEV = $(SERIAL_DEV)" >> $@
	@echo "#SERIAL_PROG = $(SERIAL_PROG)" >> $@
	@echo "#ISP_DEV = $(ISP_DEV)" >> $@
	@echo "#ISP_PROG = $(ISP_PROG)" >> $@
	@echo "#DEBUG = 1" >> $@
	@echo "created default config.mk, tune your settings there!"

-include $(PWD)/config.mk

# flags for avrdude
ifeq ($(MCU),atmega8)
	AVRDUDE_MCU=m8
endif
ifeq ($(MCU),atmega88)
	AVRDUDE_MCU=m88
endif
ifeq ($(MCU),atmega168)
	AVRDUDE_MCU=m168
endif
ifeq ($(MCU),atmega32)
	AVRDUDE_MCU=m32
endif
ifeq ($(MCU),atmega644)
	AVRDUDE_MCU=m644
endif

AVRDUDE_FLAGS += -p $(AVRDUDE_MCU) -b $(AVRDUDE_BAUDRATE) -y -F

# flags for the compiler
CFLAGS += -g -Os -finline-limit=800 -mmcu=$(MCU) -DF_CPU=$(F_CPU) -std=gnu99

# flags for the linker
LDFLAGS += -mmcu=$(MCU)

DEBUG_CFLAGS += -Wall -W -Wchar-subscripts -Wmissing-prototypes
DEBUG_CFLAGS += -Wmissing-declarations -Wredundant-decls
DEBUG_CFLAGS += -Wstrict-prototypes -Wshadow -Wbad-function-cast
DEBUG_CFLAGS += -Winline -Wpointer-arith -Wsign-compare
DEBUG_CFLAGS += -Wunreachable-code -Wdisabled-optimization
DEBUG_CFLAGS += -fshort-enums
DEBUG_CFLAGS += -Wcast-align -Wwrite-strings -Wnested-externs -Wundef
DEBUG_CFLAGS += -Wa,-adhlns=$(basename $@).lst
DEBUG_CFLAGS += -DDEBUG

ifneq ($(DEBUG),)
	CFLAGS += $(DEBUG_CFLAGS)
endif


.PHONY: sanity-check

sanity-check:
ifndef F_CPU
	@echo "please define F_CPU!"
	@exit 1
endif

ifndef MCU
	@echo "please define MCU!"
	@exit 1
endif


$(OBJECTS): sanity-check

clean:
	$(RM) *.hex *.eep.hex *.o *.lst *.lss

interactive-isp:
	$(AVRDUDE) $(AVRDUDE_FLAGS) -c $(ISP_PROG) -P $(ISP_DEV) -t

interactive-serial:
	$(AVRDUDE) $(AVRDUDE_FLAGS) -c $(SERIAL_PROG) -P $(SERIAL_DEV) -t


.PHONY: all clean interactive-isp interactive-serial

program-isp-%: %.hex
	$(AVRDUDE) $(AVRDUDE_FLAGS) -c $(ISP_PROG) -P $(ISP_DEV) -U flash:w:$<

program-isp-eeprom-%: %.eep.hex
	$(AVRDUDE) $(AVRDUDE_FLAGS) -c $(ISP_PROG) -P $(ISP_DEV) -U eeprom:w:$<

program-serial-%: %.hex
	$(AVRDUDE) $(AVRDUDE_FLAGS) -c $(SERIAL_PROG) -P $(SERIAL_DEV) -U flash:w:$<

program-serial-eeprom-%: %.eep.hex
	$(AVRDUDE) $(AVRDUDE_FLAGS) -c $(SERIAL_PROG) -P $(SERIAL_DEV) -U eeprom:w:$<

%.hex: %
	$(OBJCOPY) -O ihex -R .eeprom $< $@

%.eep.hex: %
	$(OBJCOPY) --set-section-flags=.eeprom="alloc,load" --change-section-lma .eeprom=0 -O ihex -j .eeprom $< $@

%.lss: %
	$(OBJDUMP) -h -S $< > $@

%-size: %.hex
	$(SIZE) $<


fuses:  
	$(AVRDUDE) $(AVRDUDE_FLAGS) -c $(ISP_PROG) -P $(ISP_DEV) -U lfuse:w:0xE7:m -U hfuse:w:0xDC:m 

