/* vim:fdm=marker ts=4 et ai
 * {{{
 *
 * (c) by Alexander Neumann <alexander@bumpern.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * For more information on the GPL, please go to:
 * http://www.gnu.org/copyleft/gpl.html
 }}} */

#include <avr/eeprom.h>
#include <util/crc16.h>
#include <string.h>

#include "network.h"
#include "config.h"
#include "uart.h"
#include "eeprom.h"
#include "sntp.h"
#include "syslog.h"
#include "telnetd.h"
#include "crc.h"
#include "fc.h"

#include "uip/uip.h"
#include "uip/uip_arp.h"

#ifndef ENC28J60_POLL
#define interrupt_occured() (!(INT_PIN & _BV(INT_PIN_NAME)))
#define wol_interrupt_occured() (!(WOL_PIN & _BV(WOL_PIN_NAME)))
#else
#define interrupt_occured() 0
#define wol_interrupt_occured() 0
#endif

#define BASE_CONFIG ((struct eeprom_config_base_t *)uip_buf)
#define EXT_CONFIG ((struct eeprom_config_ext_t *)uip_buf)

void network_init(void)
/* {{{ */ {

    uip_init();
    uip_arp_init();

    uip_ipaddr_t ipaddr;

    /* load base network settings */
#   ifdef DEBUG_NET_CONFIG
    uart_puts_P("net: loading base network settings\r\n");
#   endif

    /* use global network packet buffer for configuration */
    eeprom_read_block(uip_buf, EEPROM_CONFIG_BASE, sizeof(struct eeprom_config_base_t));

    /* checksum */
    uint8_t checksum = crc_checksum(uip_buf, sizeof(struct eeprom_config_base_t) - 1);

    if (checksum != BASE_CONFIG->crc) {
#       ifdef DEBUG
        uart_puts_P("net: crc mismatch: 0x");
        uart_puthexbyte(checksum);
        uart_puts_P(" != 0x");
        uart_puthexbyte(BASE_CONFIG->crc);
        uart_puts_P(" loading default settings\r\n");
#       endif

        memcpy_P(uip_ethaddr.addr, PSTR("\xAC\xDE\x48\xFD\x0F\xD0"), 6);
        uip_ipaddr(ipaddr, 10,0,0,5);
        uip_sethostaddr(ipaddr);
        uip_ipaddr(ipaddr, 255,255,255,0);
        uip_setnetmask(ipaddr);
    } else {

        /* load config settings */
        memcpy(uip_ethaddr.addr, &BASE_CONFIG->mac, 6);
        memcpy(&ipaddr, &BASE_CONFIG->ip, 4);
        uip_sethostaddr(ipaddr);
        memcpy(&ipaddr, &BASE_CONFIG->netmask, 4);
        uip_setnetmask(ipaddr);
        memcpy(&ipaddr, &BASE_CONFIG->gateway, 4);
        uip_setdraddr(ipaddr);

#       ifdef DEBUG_NET_CONFIG
        uart_puts_P("config: ip: ");
        uart_puts_ip(&uip_hostaddr);
        uart_putc('/');
        uart_puts_ip(&uip_netmask);
        uart_puts_P(" gw: ");
        uart_puts_ip(&uip_draddr);
        uart_eol();
#       endif

    }

    /* load extended network settings */
#   ifdef DEBUG_NET_CONFIG
    uart_puts_P("net: loading extended network settings\r\n");
#   endif

    /* use global network packet buffer for configuration */
    eeprom_read_block(uip_buf, (uint8_t *)EEPROM_CONFIG_EXT, sizeof(struct eeprom_config_ext_t));

    /* checksum */
    checksum = crc_checksum(uip_buf, sizeof(struct eeprom_config_ext_t) - 1);

    if (checksum != EXT_CONFIG->crc) {
#ifdef DEBUG
        uart_puts_P("net: crc mismatch: 0x");
        uart_puthexbyte(checksum);
        uart_puts_P(" != 0x");
        uart_puthexbyte(EXT_CONFIG->crc);
        uart_eol();
#endif

        memset(&sntp_server, 0, sizeof(sntp_server));
        memset(&global_syslog.server, 0, sizeof(global_syslog.server));
        global_syslog.enabled = 0;

    } else {

        memcpy(&sntp_server, &EXT_CONFIG->sntp_server, 4);
        sntp_synchronize();

        /* syslog-server */
        memcpy(&global_syslog.server, &EXT_CONFIG->syslog_server, 4);
        global_syslog.enabled = 1;

#       ifdef DEBUG_NET_CONFIG
        uart_puts_P("ext config: sntp: ");
        uart_puts_ip(&sntp_server);
        uart_puts_P(" syslog: ");
        uart_puts_ip(&global_syslog.server);
        uart_eol();
#       endif

    }

    init_enc28j60();

	telnetd_init();

} /* }}} */

void enc28j60_process_interrupts(void)
/* {{{ */ {

    /* also check packet counter, see errata #6 */
#   ifdef ENC28J60_REV4_WORKAROUND
    uint8_t pktcnt = read_control_register(REG_EPKTCNT);
#   endif

    /* if no interrupt occured and less than 5 packets are in the receive
     * buffer, return */
    if ( ! (interrupt_occured()
#   ifdef ENC28J60_REV4_WORKAROUND
                || pktcnt > 5
#   endif
           ) )
        return;

#   if defined(ENC28J60_REV4_WORKAROUND) && defined(DEBUG_REV4_WORKAROUND)
    if (pktcnt > 5)
        uart_puts_P("net: BUG: pktcnt > 5\r\n");
#   endif

    /* read interrupt register */
    uint8_t EIR = read_control_register(REG_EIR);

    /* clear global interrupt flag */
    bit_field_clear(REG_EIE, _BV(INTIE));

    /* check if some interrupts occured */
    if (EIR != 0) {

        /* check each interrupt flag the interrupt is activated for, and clear it
         * if neccessary */

#ifdef DEBUG_INTERRUPT
        uart_puts_P("net: controller interrupt, EIR = 0x");
        uart_puthexbyte(EIR);
        uart_puts_P(" = ");
        if (EIR & _BV(LINKIF))
            uart_putc('L');
        if (EIR & _BV(TXIF))
            uart_putc('T');
        if (EIR & _BV(PKTIF))
            uart_putc('P');
        if (EIR & _BV(RXERIF))
            uart_putc('r');
        if (EIR & _BV(TXERIF))
            uart_putc('t');
        uart_eol();
#endif

        /* link change flag */
        if (EIR & _BV(LINKIF)) {
            /* clear interrupt flag */
            read_phy(PHY_PHIR);

            /* read new link state */
            uint8_t link_state = (read_phy(PHY_PHSTAT2) & _BV(LSTAT)) > 0;

#ifdef DEBUG
            if (link_state) {
                uart_puts_P("net: got link!\r\n");
            } else
                uart_puts_P("net: no link!\r\n");
#endif

            if (link_state)
                syslog_message_P("etherrape booted");
        }

        /* packet transmit flag */
        if (EIR & _BV(TXIF)) {

#if 0
            /* clear send_buffer, if not waiting for ack */
            for (uint8_t i = 0; i < NET_SEND_BUFFER_SIZE; i++) {

                /* if this buffer is transmitting, remove flag */
                if (global_net.send_buffer[i].flags & _BV(SEND_BUFFER_TRANSMITTING)) {
                    global_net.send_buffer[i].flags &= ~_BV(SEND_BUFFER_TRANSMITTING);

                    /* if this buffer is not waiting for an ack, clear it */
                    if ( !(global_net.send_buffer[i].flags & _BV(SEND_BUFFER_WAIT_ACK))) {
#ifdef DEBUG_INTERRUPT
                        uart_puts_P("net: found transmitting send buffer 0x");
                        uart_puthexbyte(i);
                        uart_puts_P(", cleaning up\r\n");
#endif
                        global_net.send_buffer[i].flags = 0;
                    }
                }
            }
#endif

#ifdef DEBUG
            uint8_t ESTAT = read_control_register(REG_ESTAT);

            if (ESTAT & _BV(TXABRT))
                uart_puts_P("net: packet transmit failed\r\n");
#endif
            /* clear flags */
            bit_field_clear(REG_EIR, _BV(TXIF));
            bit_field_clear(REG_ESTAT, _BV(TXABRT) | _BV(LATECOL) );
        }

        /* packet receive flag */
        if (EIR & _BV(PKTIF)) {

            process_packet();
        }

        /* receive error */
        if (EIR & _BV(RXERIF)) {
            uart_puts_P("net: receive error!\r\n");

            bit_field_clear(REG_EIR, _BV(RXERIF));

#ifdef ENC28J60_REV4_WORKAROUND
            init_enc28j60();
#endif

        }

        /* transmit error */
        if (EIR & _BV(TXERIF)) {
#ifdef DEBUG
            uart_puts_P("net: transmit error!\r\n");
#endif

            bit_field_clear(REG_EIR, _BV(TXERIF));
        }

    }

    /* set global interrupt flag */
    bit_field_set(REG_EIE, _BV(INTIE));


} /* }}} */

void process_packet(void)
/* {{{ */ {

    /* if there is a packet to process */
    if (read_control_register(REG_EPKTCNT) == 0)
        return;

#   ifdef DEBUG_NET
    uart_puts_P("net: packet received\r\n");
#   endif

    /* read next packet pointer */
    set_read_buffer_pointer(enc28j60_next_packet_pointer);
    enc28j60_next_packet_pointer = read_buffer_memory() | (read_buffer_memory() << 8);

    /* read receive status vector */
    struct receive_packet_vector_t rpv;
    uint8_t *p = (uint8_t *)&rpv;

    for (uint8_t i = 0; i < sizeof(struct receive_packet_vector_t); i++)
        *p++ = read_buffer_memory();

    /* decrement rpv received_packet_size by 4, because the 4 byte CRC checksum is counted */
    rpv.received_packet_size -= 4;

    /* check size */
    if (rpv.received_packet_size > MAX_FRAME_LENGTH
            || rpv.received_packet_size < sizeof(struct uip_eth_hdr)
            || rpv.received_packet_size > UIP_BUFSIZE) {
#       ifdef DEBUG
        uart_puts_P("net: packet too large or too small for an ethernet header: ");
        uart_puthexbyte(HI8(rpv.received_packet_size));
        uart_puthexbyte( LO8(rpv.received_packet_size));
#       endif
        return;
    }

    /* store packet start pointer for checksum calculation */
    //global_net.packet_start_pointer = get_read_buffer_pointer();

    /* read packet */
    p = uip_buf;
    for (uint16_t i = 0; i < rpv.received_packet_size; i++)
        *p++ = read_buffer_memory();

    uip_len = rpv.received_packet_size;

    /* process packet */
    struct uip_eth_hdr *packet = (struct uip_eth_hdr *)&uip_buf;
    switch (ntohs(packet->type)) {

        /* process arp packet */
        case UIP_ETHTYPE_ARP:
#           ifdef DEBUG_NET
            uart_puts_P("net: arp packet received\r\n");
#           endif
            uip_arp_arpin();

            /* if there is a packet to send, send it now */
            if (uip_len > 0)
                transmit_packet();

            break;

        /* process ip packet */
        case UIP_ETHTYPE_IP:
#           ifdef DEBUG_NET
            uart_puts_P("net: ip packet received\r\n");
#           endif
            uip_arp_ipin();
            uip_input();

            /* if there is a packet to send, send it now */
            if (uip_len > 0) {

                /* check if an arp request has to be send */
                uip_arp_out();
                transmit_packet();
            }

            break;
#       ifdef DEBUG_UNKNOWN_PACKETS
        default:
            /* debug output */
            uart_puts_P("net: unknown packet, ");
            uart_puts_mac(&packet->src);
            uart_puts_P(" -> ");
            uart_puts_mac(&packet->dest);
            uart_puts_P(", type 0x");
            uart_puthexbyte(HI8(ntohs(packet->type)));
            uart_puthexbyte( LO8(ntohs(packet->type)));
            uart_eol();
            break;
#       endif
    }

    /* advance receive read pointer, ensuring that an odd value is programmed
     * (next_receive_packet_pointer is always even), see errata #13 */
    if ( (enc28j60_next_packet_pointer - 1) < RXBUFFER_START
            || (enc28j60_next_packet_pointer - 1) > RXBUFFER_END) {

        write_control_register(REG_ERXRDPTL, LO8(RXBUFFER_END));
        write_control_register(REG_ERXRDPTH, HI8(RXBUFFER_END));

    } else {

        write_control_register(REG_ERXRDPTL, LO8(enc28j60_next_packet_pointer - 1));
        write_control_register(REG_ERXRDPTH, HI8(enc28j60_next_packet_pointer - 1));

    }

    /* decrement packet counter */
    bit_field_set(REG_ECON2, _BV(PKTDEC));

} /* }}} */

void transmit_packet(void)
/* {{{ */ {

    /* wait for any transmits to end, with timeout */
    uint8_t timeout = 100;
    while (read_control_register(REG_ECON1) & _BV(ECON1_TXRTS) && timeout-- > 0);

    if (timeout == 0) {

#       ifdef DEBUG
        uart_puts_P("net: timeout waiting for TXRTS!\r\n");
#       endif

    } else {

        uint16_t start_pointer = TXBUFFER_START;

        /* set send control registers */
        write_control_register(REG_ETXSTL, LO8(start_pointer));
        write_control_register(REG_ETXSTH, HI8(start_pointer));

        write_control_register(REG_ETXNDL, LO8(start_pointer + uip_len));
        write_control_register(REG_ETXNDH, HI8(start_pointer + uip_len));

        /* set pointer to beginning of tx buffer */
        set_write_buffer_pointer(start_pointer);

        /* write override byte */
        write_buffer_memory(0);

        /* write data */
        for (uint16_t i = 0; i < uip_len; i++)
            write_buffer_memory(uip_buf[i]);

#       ifdef ENC28J60_REV4_WORKAROUND
        /* reset transmit hardware, see errata #12 */
        bit_field_set(REG_ECON1, _BV(ECON1_TXRST));
        bit_field_clear(REG_ECON1, _BV(ECON1_TXRST));
#       endif

        /* transmit packet */
        bit_field_set(REG_ECON1, _BV(ECON1_TXRTS));

    }

} /* }}} */

void network_handle_tcp(void)
/* {{{ */ {

#ifdef DEBUG_NET
    uart_puts_P("net_tcp: local port is 0x");
    uart_puthexbyte(HI8(uip_conn->lport));
    uart_puthexbyte(LO8(uip_conn->lport));
    uart_eol();
#endif

    if (uip_conn->lport == HTONS(TELNETD_PORT))
        telnetd_main();

} /* }}} */

void network_handle_udp(void)
/* {{{ */ {

    if (uip_udp_conn->lport == HTONS(SNTP_UDP_PORT))
        sntp_handle_conn();

    if (uip_udp_conn->lport == HTONS(SYSLOG_UDP_PORT))
        syslog_handle_conn();

    if (uip_udp_conn->lport == HTONS(FC_UDP_PORT))
        fc_handle_conn();

} /* }}} */
